/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: rectangle.C,v 1.2 1996/04/18 21:50:24 kilin Exp kilin $ 
#include <iostream.h>
#include <assert.h>
#include <stdlib.h>
#include <fstream.h>
#include "TVdefine.h"
#include "TVconstants.h"
#include "TVvector.h"
#include "TVrectangle.h"
#include "TVutil.h"

TVRectangle::TVRectangle()
{
}

TVRectangle::TVRectangle(const TVector &v, float r, int d) 
{
   int dim = v.GetDim();
   nonsig = v.ProjectFront(dim - d);
   lower = v.ProjectBack(d) - r;
   upper = v.ProjectBack(d) + r;
} 

// TVRectangle with lower, upper corners with sigdim given 
// sigdim = 0 --> Maximum non-sigdim, sig dim possible
TVRectangle::TVRectangle(const TVector& low, const TVector& high, int sigdim)
{
   int dim = min(low.GetDim(), high.GetDim());
   int firstdiff = low.FirstDiffDim(high);

   if (firstdiff == -1)
      {
	nonsig = low;
      }
   else
      {
	sigdim = (sigdim > 0 ? min(sigdim, dim - firstdiff) : dim - firstdiff);
	nonsig = low.ProjectFront(firstdiff);
  	lower = low.ProjectMid(firstdiff, sigdim);
  	upper = high.ProjectMid(firstdiff, sigdim);
      }
}


// TVRectangle with non-sig portion and sig low/high corners
TVRectangle::TVRectangle(const TVector& nsig, const TVector &low, const TVector& high)
{
  nonsig = nsig;
  if (low.GetDim() != high.GetDim())
     { 
        int d1 = min(low.GetDim(), high.GetDim());
	lower = low.ProjectFront(d1);
	upper = high.ProjectFront(d1);
     }
  else
     { 
	lower = low;
	upper = high;
     }
}

TVRectangle::TVRectangle(const TVRectangle& rec)
{
   nonsig = rec.nonsig;
   lower = rec.lower;
   upper = rec.upper;
}



TVRectangle::~TVRectangle()
{
}

TVRectangle&  TVRectangle::operator=(const TVRectangle &rec)
{
   nonsig = rec.nonsig;
   lower = rec.lower;
   upper = rec.upper;
  return *this;
}

TVector TVRectangle::GetCenter() const
{
  return (nonsig  << ((lower + upper) / 2));
}

float TVRectangle::GetRadius() const
{
  return (upper - lower).MaxComponent();
}

int TVRectangle::GetSigDim() const
{
  return lower.GetDim();
}

int TVRectangle::GetNonSigDim() const
{
  return nonsig.GetDim();
}

int TVRectangle::GetCenterDim() const
{
  return nonsig.GetDim() + lower.GetDim();
}


TVector TVRectangle::GetSigCenter() const
{
  return (lower + upper) / 2;
}

TVector TVRectangle::GetNonSigCenter() const
{
  return nonsig;
}


int TVRectangle::Size() const
{
   return  nonsig.Size() + lower.Size() + upper.Size();
}


TVRectangle& TVRectangle::PutNonSig(const TVector& v)
{
  nonsig = v;
  return *this;
}
   
TVRectangle& TVRectangle::PutSig(const TVector& l, const TVector& h)
{
  lower = l;
  upper = h;
  return *this;
}

TVRectangle& TVRectangle::Put(const TVector& ns, const TVector& l, const TVector& h)
{
  nonsig = ns;
  lower = l;
  upper = h;
  return *this;
}

TVRectangle& TVRectangle::AppendSig(const TVector& l, const TVector& h)
{
  assert (l.GetDim() == h.GetDim());

  lower <<= l;
  upper <<= h;

  return *this;
}


TVRectangle TVRectangle::ProjectSigTVRectangle() const
{
  TVRectangle res;

  res.PutSig(lower, upper);

  return(res);
}

TVRectangle TVRectangle::ProjectFront(int dim) const
{
  assert(dim <= GetCenterDim());
  TVRectangle res;
  int sig = dim - nonsig.GetDim();

  if (sig <= 0)
     res.PutNonSig(nonsig.ProjectFront(dim));
  else
     res.Put(nonsig, lower.ProjectFront(sig), upper.ProjectFront(sig));

/*
cout << " res  : " << res;
cout << (dim - GetNonSigDim() > 0 ? dim - GetNonSigDim() : SIG_DIM) << '\n';
*/
  return(res);
}

TVRectangle TVRectangle::ProjectMid(int start, int dim) const
{
  assert(start + dim <= GetCenterDim());
  TVRectangle res;
  int sig = dim + start - nonsig.GetDim();
  
  if (sig <= 0)
     res.PutNonSig(nonsig.ProjectMid(start, dim));
  else
     res.Put(nonsig.ProjectBack(nonsig.GetDim() - start + 1), lower.ProjectFront(sig), upper.ProjectFront(sig));
  
  return(res);
}

TVRectangle TVRectangle::ProjectBack(int dim) const
{
  assert(dim <= GetCenterDim());
  TVRectangle res;
  int nonsigdim = dim - lower.GetDim();

  if (dim < 0)
     res.PutSig(lower.ProjectBack(dim), upper.ProjectBack(dim));
  else 
     res.Put(nonsig.ProjectBack(nonsigdim), lower, upper);
  

  return(res);
}

TVector TVRectangle::ProjectFrontLow(int dim) const
{
   if (dim <= nonsig.GetDim())
     return nonsig.ProjectFront(dim);

   if (!nonsig.GetDim())
     return lower.ProjectFront(dim);

   return nonsig << lower.ProjectFront(dim - nonsig.GetDim());
}

TVector TVRectangle::ProjectFrontHigh(int dim) const
{
   if (dim <= nonsig.GetDim())
     return nonsig.ProjectFront(dim);

   if (!nonsig.GetDim())
     return lower.ProjectFront(dim);

   return nonsig << upper.ProjectFront(dim - nonsig.GetDim());
}

TVector TVRectangle::ProjectMidLow(int start, int dim) const
{
   if (start + dim <= nonsig.GetDim())  
     return nonsig.ProjectMid(start, dim);

   if (!nonsig.GetDim())
       return lower.ProjectMid(start, dim);

   return nonsig.ProjectBack(GetNonSigDim() - start) << lower.ProjectFront(dim - GetNonSigDim() + start);
}

TVector TVRectangle::ProjectMidHigh(int start, int dim) const
{
   if (start + dim <= nonsig.GetDim())  
     return nonsig.ProjectMid(start, dim);

   if (!nonsig.GetDim())
       return upper.ProjectMid(start, dim);

   return nonsig.ProjectBack(GetNonSigDim() - start) << upper.ProjectFront(dim - GetNonSigDim() + start);
}



int TVRectangle::Contain(const TVector& v) const
{
   int result = FALSE;
   
   if (v.GetDim() >= GetCenterDim())
      {
	  if (GetNonSigDim())
             result = (v.ProjectFront(GetNonSigDim()) == nonsig) &&  
		      (v.ProjectMid(GetNonSigDim(), GetSigDim()).Inrange(lower, upper));
          else
	     result = (v.ProjectFront(GetSigDim()).Inrange(lower, upper));
      }
   return result;
}

int TVRectangle::Contain(const TVRectangle &d) const
{
   int part1 = TRUE, part2 = TRUE, part3 = TRUE;

   if ((d.GetCenterDim() < GetCenterDim()) || (d.GetNonSigDim() < GetNonSigDim()))
      return FALSE;	// Inner rectangle must have more dimensions pinned down
			// Both in sig. and total dimensions
   else
      {
       //    Algorithm
       //
       //	     in  : |---------------------------|-----------------|
	//		         non-sig  |            |    sig
	//				  |            |
	//			          |            |
	//	     out : |--------------|---------------------|
	//		        non-sig                    sig
	//		       (1)           (2)         (3)

         // Check part (1)
	 if (GetNonSigDim())
	    part1 = (nonsig == d.nonsig.ProjectFront(GetNonSigDim()));

         // Check part (2)

	 int surplus_nonsigdim = min(d.GetNonSigDim() - GetNonSigDim(), GetSigDim());
         if (surplus_nonsigdim)
	    part2 = d.nonsig.ProjectMid(GetNonSigDim(), surplus_nonsigdim).Inrange(lower.ProjectFront(surplus_nonsigdim), upper.ProjectFront(surplus_nonsigdim));

         // Check part (3)

	 int part3dim = GetCenterDim() - d.GetNonSigDim();
         if (part3dim > 0)
	    {
	        TVector dlower = d.lower.ProjectFront(part3dim);
	        TVector dupper = d.upper.ProjectFront(part3dim);

 		TVector vlower = lower.ProjectMid(surplus_nonsigdim, part3dim);
 		TVector vupper = upper.ProjectMid(surplus_nonsigdim, part3dim);

                part3 = dlower.Inrange(vlower, vupper) && dupper.Inrange(vlower, vupper);
            }
        }

    return part1 && part2 && part3;
}

int TVRectangle::Intersect(const TVRectangle& d) const
{
   TVRectangle d1, d2;

   // ensure d1 has the shorter non-sigdim
   if (GetNonSigDim() < d.GetNonSigDim())
      {
	 d1 = *this;
	 d2 = d;
      }
   else
      {
	 d2 = *this;
	 d1 = d;
      }

   int part1 = TRUE, part2 = TRUE, part3 = TRUE;

   // Check part (1)
   if (d1.GetNonSigDim())
      part1 = (d1.nonsig == d2.nonsig.ProjectFront(d1.GetNonSigDim()));

   // Check part (2)

   int surplus_nonsigdim = min(d2.GetNonSigDim() - d1.GetNonSigDim(), d1.GetSigDim());
   if (surplus_nonsigdim)
      part2 = d2.nonsig.ProjectMid(d1.GetNonSigDim(), surplus_nonsigdim).Inrange(d1.lower.ProjectFront(surplus_nonsigdim), d1.upper.ProjectFront(surplus_nonsigdim));


   // Check part 3
   int restdim = min(d2.GetSigDim(), d1.GetSigDim() - surplus_nonsigdim);
   int restdimstart =  surplus_nonsigdim;
   if (restdim > 0)
      {
	int i = 0;
        for (; (i < restdim) && intersect((d1.lower)[restdimstart + i], 
                                          (d1.upper)[restdimstart + i], 
                                          (d2.lower)[i], 
                                          (d2.upper)[i]); i++)
            ;

        part3 = (i == restdim);
      }

// cout << *this << " Intersect " << d << " : " << (part1 && part2 && part3 ? "TRUE" : "FALSE") << endl;
   return part1 && part2 && part3;


}

int TVRectangle::NoIntersectDim(const TVRectangle& rect) const
{
   TVRectangle d1, d2;
   int result = 0;
   int curdim;

   // ensure d1 has the shorter non-sigdim
   if (GetNonSigDim() < rect.GetNonSigDim())
      {
	 d1 = *this;
	 d2 = rect;
      }
   else
      {
	 d2 = *this;
	 d1 = rect;
      }

   int shortnonsig = d1.GetNonSigDim(), longnonsig = d2.GetNonSigDim();
   for (curdim = 0; curdim < shortnonsig; curdim++)
      {
         result += (d1.nonsig[curdim] != d2.nonsig[curdim]);
      }

   for (; curdim < longnonsig; curdim++)
      result += ((d2.nonsig[curdim] < d1.lower[curdim - shortnonsig]) ||
		 (d2.nonsig[curdim] > d1.upper[curdim - shortnonsig]));

   int mdim = min(d1.GetCenterDim(), d2.GetCenterDim());
   for (; curdim < mdim; curdim++)
      result += !intersect(d1.lower[curdim - shortnonsig], d1.upper[curdim - shortnonsig],
			   d2.lower[curdim - longnonsig], d2.upper[curdim - longnonsig]);

   return result;
}

// Assume all parts are significiant
void NewSigRect(TVector &lower, TVector& upper, const TVector& oldlower, const TVector& oldupper, const TVector& v, int sigdim = 0)
{
   int dim = (sigdim ? min(oldlower.GetDim(), sigdim) : oldlower.GetDim());

   lower = oldlower.ProjectFront(dim).MinCompo(v.ProjectFront(dim));
   upper = oldupper.ProjectFront(dim).MaxCompo(v.ProjectFront(dim));

}


TVRectangle TVRectangle::InsertTVector(const TVector& v, int result_sig_dim)
{
  TVRectangle result;
  
  if (!Contain(v))
     {
        if (GetNonSigDim() >= v.GetDim())
           {
	     // TVector shorter than nonsig portion
	     return NewRect(nonsig, v, result_sig_dim);
           }
	else
	   {
	     if (!(GetNonSigDim()))
		{
	           // No nonsig portion
                   NewSigRect(result.lower, result.upper, lower, upper, v, result_sig_dim); 
		}
	     else
		{
		  int min_nonequal = nonsig.FirstDiffDim(v);
	     	  if (min_nonequal == -1)
		     {
		       // TVector have same value at all nonsig portions
		       result.nonsig = nonsig;
                       NewSigRect(result.lower, result.upper, lower, upper, v.ProjectBack(v.GetDim() - GetNonSigDim()), result_sig_dim); 
		     }
		  else
		     {
		       result.nonsig = nonsig.ProjectFront(min_nonequal);
		       int mindim = min(v.GetDim(), GetCenterDim());

		       // calculate the real non sig portion	
		       if (!result_sig_dim)
			  result_sig_dim = mindim - min_nonequal;
		       else
			  if (min_nonequal + result_sig_dim > mindim)
			     result_sig_dim = mindim - min_nonequal;
			  

		       if (min_nonequal + result_sig_dim <= GetNonSigDim())
			  {
		             TVector v1 = nonsig.ProjectMid(min_nonequal, result_sig_dim);
		             TVector v2 = v.ProjectMid(min_nonequal, result_sig_dim);
			     result.lower = v1.MinCompo(v2);
			     result.upper = v1.MaxCompo(v2);
			  } 
		       else
			  {
		             TVector onon = nonsig.ProjectBack(GetNonSigDim() - min_nonequal);
		             TVector v1 = v.ProjectMid(min_nonequal, onon.GetDim());
			          
		             TVector ol = lower.ProjectFront(result_sig_dim - onon.GetDim());
		             TVector ou = upper.ProjectFront(result_sig_dim - onon.GetDim());
		             TVector v2 = v.ProjectMid(GetNonSigDim(),  result_sig_dim - onon.GetDim());
	
			     TVector nl, nu;
			 
                             NewSigRect(nl, nu, ol, ou, v2);

			     result.lower = onon.MinCompo(v1) << nl;
			     result.upper = onon.MaxCompo(v1) << nu;

			  }

		     }
		}
	   }
       return result;
     }
  else
    {
     return *this;
    }
}



TVRectangle TVRectangle::Combine(const TVRectangle& d, int result_sig_dim)
{
  const TVRectangle *dshort, *dlong;
  TVRectangle result;

       //    Algorithm
       //
       //	    dlong : |----------------------------------------|-----------------|
	//		          non-sig   |           |            |    sig
	//			    	    |           |            |
	//			            |           |            |
	//	   dshort : |---------------|-----------|---------------------|
	//		        non-sig                    sig
	//		         (1)             (1a)          (2)         (3)
        //		       <equal>	     <unequal>

  if (GetNonSigDim() > d.GetNonSigDim())
     {
        dshort = &d;
	dlong = this;
     }
  else 
     {
        dshort = this;
	dlong = &d;
     }

  int nonsigshort = dshort->GetNonSigDim();

  int nonsig_firstdiff = nonsig.FirstDiffDim(d.nonsig); 

  int cursigdim = 0;
  int dim1a = 0, dim2 = 0, dim3 = 0;

  // Case (1)
  if (nonsig_firstdiff == -1)
     {
	// Equal nonsigdim
	result.nonsig = dshort->nonsig; 
     }
  else
     {
       if (nonsig_firstdiff)
           result.nonsig = dshort->nonsig.ProjectFront(nonsig_firstdiff); 
     }


  // Case (1a)

  if (nonsig_firstdiff >= 0)
     {
	dim1a = nonsigshort - nonsig_firstdiff; 
	if ((result_sig_dim > 0) && (dim1a > result_sig_dim))
	    dim1a = result_sig_dim;

	result.lower = dshort->nonsig.ProjectMid(nonsig_firstdiff, dim1a).MinCompo(dlong->nonsig.ProjectMid(nonsig_firstdiff, dim1a)); 
	result.upper = dshort->nonsig.ProjectMid(nonsig_firstdiff, dim1a).MaxCompo(dlong->nonsig.ProjectMid(nonsig_firstdiff, dim1a)); 
	cursigdim = dim1a;
     }

  int nonsigdiff = dlong->GetNonSigDim() - nonsigshort;
  // case (2)

  if ( (!result_sig_dim) || (cursigdim < result_sig_dim))
     {
	 if (nonsigdiff)
	    {
		dim2 = nonsigdiff;
	        if ((result_sig_dim > 0) && (dim2 + cursigdim > result_sig_dim))
	            dim2 = result_sig_dim - cursigdim;
		result.lower <<= dlong->nonsig.ProjectMid(nonsigshort, dim2).MinCompo(dshort->lower.ProjectFront(dim2));
		result.upper <<= dlong->nonsig.ProjectMid(nonsigshort, dim2).MaxCompo(dshort->upper.ProjectFront(dim2));
		cursigdim += dim2;
	    }
     }

  // case (3)

  if ( (!result_sig_dim) || (cursigdim < result_sig_dim))
     {
	dim3 = dshort->GetCenterDim() - dlong->GetNonSigDim();
	if (dim3)
	   {
	      dim3 = min(dim3, dlong->GetSigDim());	
	      if ((result_sig_dim > 0) && (dim3 + cursigdim > result_sig_dim))
	            dim3 = result_sig_dim - cursigdim;

		
	      result.lower <<= dlong->lower.ProjectFront(dim3).MinCompo(dshort->lower.ProjectMid(nonsigdiff, dim3));
	      result.upper <<= dlong->upper.ProjectFront(dim3).MaxCompo(dshort->upper.ProjectMid(nonsigdiff, dim3));

	   }
     }


  return result;
}




TVRectangle operator<<(const TVector& v , const TVRectangle& d)
{
   TVRectangle res(v << d.nonsig, d.lower, d.upper);

   return res;
}



int sort_nonsig(const void *d1, const void *d2)
{
  TVRectangle *r1 = (TVRectangle *)d1;
  TVRectangle *r2 = (TVRectangle *)d2;
  
  int n1 = r1->GetNonSigDim();
  int n2 = r2->GetNonSigDim();

  return(n1 - n2);
}


TVRectangle MinBoundTVRectangle(TVRectangle *darray, int count, int res_sig_dim)
{

/*
cout << "Enter minboundrect(rect)\n";
for (int xyy = 0; xyy < count; xyy++)
    cout << xyy << " : " << darray[xyy] << endl;
*/

  assert(count > 0);
  TVRectangle result;
   
  if (count == 1)
     return darray[0];
  
  TVRectangle *darray1 = new TVRectangle[count];
  int i;
  for (i = 0; i < count ; i++)
      darray1[i] = darray[i];
  qsort(darray1, count, sizeof(TVRectangle), sort_nonsig);

  int min_nonsigdim = darray1[0].GetNonSigDim();
  int min_nonequal = -1;
  int min_dim = darray1[0].GetCenterDim();

  for (i = 1; i < count; i++)
    {
       int new_minnonequal = darray1[i-1].GetNonSigCenter().FirstDiffDim(darray1[i].GetNonSigCenter());
       if (min_nonequal == -1)
	  min_nonequal = new_minnonequal;
       else
	  if (new_minnonequal != -1)
	     min_nonequal = min(min_nonequal, new_minnonequal);
       min_dim = min(min_dim, darray1[i].GetCenterDim());

    }

  // Calculate nonsig

  if ((min_nonequal == -1) || (min_nonequal >= min_nonsigdim))
     {
        result.PutNonSig(darray1[0].GetNonSigCenter());
     }
  else
     {
	if (min_nonequal)
	   result.PutNonSig(darray1[0].GetNonSigCenter().ProjectFront(min_nonequal));
     }

  int res_nonsigdim = result.GetNonSigDim();

  // Calculate min dim
  if ((!res_sig_dim) || (res_sig_dim > min_dim))
     {
	res_sig_dim = min_dim;
     }

  result.lower = darray1[0].ProjectMidLow(res_nonsigdim, res_sig_dim);
  result.upper = darray1[0].ProjectMidHigh(res_nonsigdim, res_sig_dim);

  for (i = 1; i < count; i++)
     {
       result.lower = result.lower.MinCompo(darray1[i].ProjectMidLow(res_nonsigdim, res_sig_dim));
       result.upper = result.upper.MaxCompo(darray1[i].ProjectMidHigh(res_nonsigdim, res_sig_dim));
     }
  
  delete [] darray1;
// cout << "Result : " << result << endl;
  return result;
}


TVRectangle MinBoundTVRectangle(TVector *varray, int count, int res_sig_dim)
{
  assert(count > 0);
  TVRectangle result;
   
  if (count == 1)
     {
	result.PutNonSig(varray[0]);
        return result;
     }
  
  int min_nonequal = -1;
  int min_dim = varray[0].GetDim();
  for (int i = 1; i < count; i++)
    {
       int new_minnonequal =  varray[i-1].FirstDiffDim(varray[i]);
       if (min_nonequal == -1) 
	  min_nonequal = new_minnonequal;
       else
	  if (new_minnonequal != -1)
	     min_nonequal = min(min_nonequal, new_minnonequal);
       min_dim = min(min_dim, varray[i].GetDim());

    }
  if (min_nonequal == -1)  
     {
	// all vectors agress in the dimensions
	// create a 0 radius rectangle with specified sig_dim
	result.PutNonSig(varray[0].ProjectFront(min_dim));
     }
  else
    if (min_nonequal >= min_dim)
     {
	result.PutNonSig(varray[0].ProjectFront(min_dim));
     }
    else 
     {
	result.PutNonSig(varray[0].ProjectFront(min_nonequal));

	// determine non-sig-dim & sig-dim; 
	int sigdim;
        if (res_sig_dim)
	   sigdim = min(min_dim - min_nonequal, res_sig_dim);   // can consider just unfold to min_dim
	else
 	   sigdim = min_dim - min_nonequal;
	// MinBoundTVRectangle2 for vectors

        result.lower = varray[0].ProjectMid(min_nonequal, sigdim);
        result.upper = varray[0].ProjectMid(min_nonequal, sigdim);
        for (int i = 1; i < count; i++)
          {
               result.lower = result.lower.MinCompo(varray[i].ProjectMid(min_nonequal, sigdim));
               result.upper = result.upper.MaxCompo(varray[i].ProjectMid(min_nonequal, sigdim));
          }
     }

  return result;
}

TVRectangle NewRect(const TVector& v1, const TVector& v2, int maxsigdim)
{
   TVRectangle res;
   int mindim = min(v1.GetDim(), v2.GetDim());
   int firstdiff = v1.FirstDiffDim(v2);

   if (firstdiff > -1)   
      {
	 // sig dim exists
	if (firstdiff > 0)
	   res.PutNonSig(v1.ProjectFront(firstdiff));
	if ((!maxsigdim) || ((mindim - firstdiff) < maxsigdim))
	   maxsigdim = mindim - firstdiff;
  	TVector v1a = v1.ProjectMid(firstdiff, maxsigdim); 	
  	TVector v2a = v1.ProjectMid(firstdiff, maxsigdim); 	
	
        res.PutSig(v1a.MinCompo(v2a), v1a.MaxCompo(v2a));
      } 
   else
      res.PutNonSig(v1.ProjectFront(mindim));
   
   return res;
}


double TVRectangle::MinDist(const TVector& v) const
{
  assert(v.GetDim() >= GetCenterDim());

  double result = 0.0;

  if (nonsig.GetDim())
     result += nonsig.Distance(v.ProjectFront(nonsig.GetDim()));

  for (int i = 0; i < lower.GetDim(); i++)
      {
        VCOM_TYPE compo = v[nonsig.GetDim() + i];
        if (compo < lower[i])
	   result += (lower[i] - compo);
        else
           if (compo > upper[i])
	    result += (compo - upper[i]);
      }

  return result; 
}




//  befriended functions

ostream& operator<< (ostream& os, const TVRectangle& d)
{
   os << " NonSig : " <<  d.nonsig << "  SigLower : " << d.lower << "  Sig Upper : " << d.upper;
   return os;
}

ostream& operator<< (ostream& os, const TVRectangle*& d)
{
   os << " NonSig : " <<  d->nonsig << "  SigLower : " << d->lower << "  Sig Upper : " << d->upper;
   return os;
}

ofstream& operator< (ofstream& os, const TVRectangle& d)
{
   os < d.nonsig;
   os << " ";
   os < d.lower;
   os << " ";
   os < d.upper; 
   os << " ";
   return os;
}

ofstream& operator< (ofstream& os, const TVRectangle*& d)
{
   os < d->nonsig;
   os << " ";
   os < d->lower;
   os << " ";
   os < d->upper; 
   os << " ";
   return os;
}

istream& operator>> (istream& is, TVRectangle& d)
{
   cout << "Input NonSig : ";
   is >> d.nonsig;
   cout << "Input SigLower : ";
   is >> d.lower;
   cout << "Input SigUpper : ";
   is >> d.upper;
   return is;
}

istream& operator>> (istream& is, TVRectangle*& d)
{
   cout << "Input NonSig : ";
   is >> d->nonsig;
   cout << "Input SigLower : ";
   is >> d->lower;
   cout << "Input SigUpper : ";
   is >> d->upper;
   return is;
}


ifstream& operator> (ifstream& ifs, TVRectangle& d)
{
   ifs > d.nonsig;
   ifs > d.lower;
   ifs > d.upper;
   return ifs;
}

ifstream& operator> (ifstream& ifs, TVRectangle*& d)
{
   ifs > d->nonsig;
   ifs > d->lower;
   ifs > d->upper;
   return ifs;
}


